@@ -62,9 +62,9 @@ class PerturbFlow(nn.Module):
                  supervised_mode: bool = False,
                  z_dim: int = 10,
                  z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'normal',
-                 loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'negbinomial',
+                 loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'multinomial',
                  inverse_dispersion: float = 10.0,
-                 use_zeroinflate: bool = True,
+                 use_zeroinflate: bool = False,
                  hidden_layers: list = [300],
                  hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
                  nn_dropout: float = 0.1,
@@ -226,14 +226,14 @@ class PerturbFlow(nn.Module):
                     )
             
         self.decoder_concentrate = MLP(
-                [self.latent_dim] + self.decoder_hidden_layers + [self.input_size],
-                activation=activate_fct,
-                output_activation=None,
-                post_layer_fct=post_layer_fct,
-                post_act_fct=post_act_fct,
-                allow_broadcast=self.allow_broadcast,
-                use_cuda=self.use_cuda,
-            )
+                    [self.latent_dim] + self.decoder_hidden_layers + [self.input_size],
+                    activation=activate_fct,
+                    output_activation=None,
+                    post_layer_fct=post_layer_fct,
+                    post_act_fct=post_act_fct,
+                    allow_broadcast=self.allow_broadcast,
+                    use_cuda=self.use_cuda,
+                )
 
         if self.latent_dist == 'studentt':
             self.codebook = MLP(
@@ -322,7 +322,7 @@ class PerturbFlow(nn.Module):
             gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
             
         acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)
-
+        
         I = torch.eye(self.code_size)
         if self.latent_dist=='studentt':
             acs_dof,acs_loc = self.codebook(I)
@@ -351,12 +351,13 @@ class PerturbFlow(nn.Module):
 
             zs = zns
             concentrate = self.decoder_concentrate(zs)
-            if self.loss_func == 'bernoulli':
+            if self.loss_func in ['bernoulli']:
                 log_theta = concentrate
             else:
                 rate = concentrate.exp()
-                if self.loss_func != 'poisson':
-                    theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                if self.loss_func == 'poisson':
+                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
                 if self.use_zeroinflate:
@@ -428,24 +429,19 @@ class PerturbFlow(nn.Module):
                 zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
 
             if self.cell_factor_size>0:
-                #zus = None
-                #for i in np.arange(self.cell_factor_size):
-                #    if i==0:
-                #        zus = self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
-                #    else:
-                #        zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
                 zus = self._total_effects(zns, us)
                 zs = zns+zus
             else:
                 zs = zns
 
             concentrate = self.decoder_concentrate(zs)
-            if self.loss_func == 'bernoulli':
+            if self.loss_func in ['bernoulli']:
                 log_theta = concentrate
             else:
                 rate = concentrate.exp()
-                if self.loss_func != 'poisson':
-                    theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                if self.loss_func == 'poisson':
+                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
                 if self.use_zeroinflate:
@@ -535,12 +531,13 @@ class PerturbFlow(nn.Module):
             zs = zns
 
             concentrate = self.decoder_concentrate(zs)
-            if self.loss_func == 'bernoulli':
+            if self.loss_func in ['bernoulli']:
                 log_theta = concentrate
             else:
                 rate = concentrate.exp()
-                if self.loss_func != 'poisson':
-                    theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                if self.loss_func == 'poisson':
+                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
                 if self.use_zeroinflate:
@@ -566,6 +563,8 @@ class PerturbFlow(nn.Module):
                 #zn_loc, zn_scale = self.encoder_zn(xs)
                 zn_loc, zn_scale = self._get_basal_embedding(xs)
                 zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
+            else:
+                zns = embeds
 
     def model4(self, xs, us, ys, embeds=None):
         pyro.module('PerturbFlow', self)
@@ -638,12 +637,13 @@ class PerturbFlow(nn.Module):
                 zs = zns
 
             concentrate = self.decoder_concentrate(zs)
-            if self.loss_func == 'bernoulli':
+            if self.loss_func in ['bernoulli']:
                 log_theta = concentrate
             else:
                 rate = concentrate.exp()
-                if self.loss_func != 'poisson':
-                    theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+                if self.loss_func == 'poisson':
+                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
                 if self.use_zeroinflate:
@@ -669,6 +669,8 @@ class PerturbFlow(nn.Module):
                 #zn_loc, zn_scale = self.encoder_zn(xs)
                 zn_loc, zn_scale = self._get_basal_embedding(xs)
                 zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
+            else:
+                zns = embeds
 
     def _total_effects(self, zns, us):
         zus = None
@@ -870,11 +872,27 @@ class PerturbFlow(nn.Module):
         R = np.concatenate(R)
         return R
     
-    def _count(self,concentrate):
+    def _count(self,concentrate, library_size=None):
         if self.loss_func == 'bernoulli':
-            counts = self.sigmoid(concentrate)
-        else:
-            counts = concentrate.exp()
+            #counts = self.sigmoid(concentrate)
+            counts = dist.Bernoulli(logits=concentrate).to_event(1).mean
+        elif self.loss_func == 'negbinomial':
+            #counts = concentrate.exp()
+            #rate = concentrate.exp()
+            #theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+            
+            total_count = pyro.param("inverse_dispersion")
+            counts = dist.NegativeBinomial(total_count=total_count, logits=concentrate).to_event(1).mean
+        elif self.loss_func == 'poisson':
+            rate = concentrate.exp()
+            theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+            rate = theta * library_size
+            counts = dist.Poisson(rate=rate).to_event(1).mean
+        elif self.loss_func == 'multinomial':
+            rate = concentrate.exp()
+            theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
+            #counts = dist.Multinomial(total_count=int(1e8), probs=theta).mean
+            counts = theta * library_size
         return counts
     
     def _count_sample(self,concentrate):
@@ -886,22 +904,36 @@ class PerturbFlow(nn.Module):
             counts = dist.Poisson(rate=counts).to_event(1).sample()
         return counts
     
-    def get_counts(self, zs, 
+    def get_counts(self, zs, library_sizes = None,
                         batch_size: int = 1024, 
                         use_sampler: bool = False):
 
         zs = convert_to_tensor(zs, device=self.get_device())
-        dataset = CustomDataset(zs)
+        
+        if self.loss_func in ['multinomial','poisson']:
+            assert library_sizes is not None, 'Library sizes are required for multinomial!'
+            
+            if type(library_sizes) == list:
+                library_sizes = np.array(library_sizes).view(-1,1)
+            elif len(library_sizes.shape)==1:
+                library_sizes = library_sizes.view(-1,1)
+            ls = convert_to_tensor(library_sizes, device=self.get_device())
+        else:
+            ls = zs
+            
+        dataset = CustomDataset2(zs,ls)
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
         
         E = []
         with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
-            for Z_batch, _ in dataloader:
+            for Z_batch, L_batch, _ in dataloader:
+                if self.loss_func != 'multinomial':
+                    L_batch = None
                 concentrate = self._get_expression_response(Z_batch)
                 if use_sampler:
                     counts = self._count_sample(concentrate)
                 else:
-                    counts = self._count(concentrate)
+                    counts = self._count(concentrate, L_batch)
                 E.append(tensor_to_numpy(counts))
                 pbar.update(1)
         
@@ -933,7 +965,7 @@ class PerturbFlow(nn.Module):
             decay_rate: float = 0.9,
             config_enum: str = 'parallel',
             threshold: int = 0,
-            use_jax: bool = False):
+            use_jax: bool = True):
         """
         Train the PerturbFlow model.
 
